/*
**  CXSC is a C++ library for eXtended Scientific Computing (V 2.0_beta3)
**
**  Copyright (C) 1990-2000 Institut fuer Angewandte Mathematik,
**                          Universitaet Karlsruhe, Germany
**            (C) 2000-2003 Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany
**
**  This program 'Min-Incl' is free software; you can redistribute it 
**  and/or modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
****************************************************************************
**  The program    'Min-Incl'    calculates an inclusion [x1,x2] of the   **
**  the minimum point x0 and an inclusion [y1,y2]  of the minimum value   **
**  f(x0) of a function  f ,which is convex and has a relative minimum.   **
**  Furthermore [y1,y2] is an inclusion of f([x1,x2]).                    **
**  Kraemer|Blomquist, 25.01.05;                                          **
****************************************************************************
*/

#include <iostream>      // For cout, cerr, ...
#include <interval.hpp>
#include <rmath.hpp>  

using namespace std;
using namespace cxsc;

// *************************************************************************
// *****     The real convex function   f(x)  must be defined here:     ****
// *************************************************************************
real f(const real& x)
{   // Definition of the convex function f(x) 
    real y;
    y = (x>0) ? x*x*ln(x) : x*x;
    return y;
}
// *************************************************************************

void SCHRANKE(const real& x,    // function argument
              const real& eps,  // Relative error bound of the function
	      real& fu,         // Lower bound of the exact value f(x)
	      real& fo)         // Upper bound of the exact value f(x)
{
    real amin, aplus, fw;
    int k;
    amin  = subd(1.0,eps);
    aplus = addu(1.0,eps);
    fw = f(x);

    if (fw==0) k=0;
    if (fw>0)  k=1;
    if (fw<0)  k=2;
    switch(k)
    {
	case 0:
	    fu=0; fo=0;
	    break;
	case 1:
	    fu = divd(fw,aplus);
	    fo = divu(fw,amin);
	    break;
	case 2:
	    fu = divd(fw,amin);
	    fo = divu(fw,aplus);
	    break;
    }
} // SCHRANKE

void X_interval(const real& eps, // Relative error bound of f(x)
                const real& ZP,  // Factor determing the step size
                real& xs,        // For a given approximation xs of x0 
                                 // an inclusion [xa,xb] is calculated
		real& xa,        
		real& xb)        // xs=0.5*(xa+xb) is a new approximation
/*  For a first approximation xs of x0 an inclusion interval [xa,xb]   */
/*  for the minimum point xo is calculated.       The returned value   */ 
/*  xs := 0.5*(xa+xb) is then an improved new approximation of x0.     */
{
    real OS,fu,fo,x,delta,xalt;
    int n;
    SCHRANKE(xs,eps,fu,OS);      // fu < f(xs) < OS;
    x = xs;  delta = abs(x*ZP);
    n = 0;
    do {
	xalt = x;
	x = subd(x,delta);  // running to the left side
	SCHRANKE(x,eps,fu,fo);
	n++;
	if (n==10) {n = 0;  delta *= 1.5;}
    } while (fu<OS);
    delta /= 200;  // refinement of the left value xalt
    while (xalt == xalt-delta) delta *= 1.5;

    do {
	xalt = subd(xalt,delta);
	SCHRANKE(xalt,eps,fu,fo);
	delta *= 1.5;
    } while (fu<=OS);
    xa = xalt; // xa < x0
    x = xs;  delta = abs(x*ZP);  n = 0;
    do {
	xalt = x;
	x = addu(x,delta);  // running to the right side
	SCHRANKE(x,eps,fu,fo);
	n++;
	if (n==10) {n = 0; delta *= 1.5;}
    } while (fu<OS);
    delta /= 200;  // refinement of the right value xalt
    while (xalt == xalt+delta) delta *= 1.5;

    do {
	xalt = addu(xalt,delta);
	SCHRANKE(xalt,eps,fu,fo);
	delta *= 1.5;
    } while (fu<=OS);
    xb = xalt;  // xa < x0 < xb
    xs = (xa+xb) / 2;  // xs: A new approximation for x0
} // X_interval

void X_itera(const real& eps,
	     real& ZP,
	     real& xs,
	     real& xa,
	     real& xb)
/*  xs: Approximation of the exact minimum point x0.            */
/*  ZP = 10^(-2); |xs*ZP| is the initial step size.             */
/*  [xa,xb] is an inclusion of x0; several iteration are done.  */
{
    real x1,x2,x1alt,x2alt,lalt,lneu;
    x1 = -1e306;  x2 = -x1;  lneu = x2 - x1;
    do {
	x1alt = x1;  x2alt = x2;  lalt = lneu;
	X_interval(eps,ZP,xs,x1,x2);
	ZP /= 16;  lneu = x2 - x1;
    } while (lneu<lalt);
    xa = x1alt;   xb = x2alt;
} // X_itera

void MIN_EINSCHL(const real& eps,
		 real& ZP,
		 real& xs,
		 real& x1,
		 real& x2,
		 real& y1,
		 real& y2)
{
    real xa,xb,deltax,x,y,y1l,y2l,m1,m2,m,m1alt,b1,b2,c1,c2;
    int n;

    X_itera(eps,ZP,xs,xa,xb);  // xa < x0 < xb
    // Calculation of the right secant equation:  y=m1*x+b1, m1>0;
    deltax = (xb-xa) / 100;  m1 = 1e306;
    SCHRANKE(xb,eps,y1l,y);
    do {
	m1alt = m1;
	deltax *= 1.1;
	x = addu(xb,deltax); // running to the right to decrease
        // the secant slope until this slope will increase again.
	SCHRANKE(x,eps,y,y2l);
	c1 = subu(y2l,y1l);  
	c2 = subd(x,xb);
	m1 = divu(c1,c2);    // m1: positive secant slope
    } while (m1<=m1alt);
    deltax /= 1.21;  // Going two step backwards; starting from
    // this point with a smaller step size to the right for getting
    // a more smaller secant slope.
    m = m1alt;   m1 = 1e306;
    do {
	m1alt = m1;
	deltax *= 1.001;
	x = addu(xb,deltax); // The secant slope should be as small
                             // as possible! 
	SCHRANKE(x,eps,y,y2l);
	c1 = subu(y2l,y1l);  
	c2 = subd(x,xb);
	m1 = divu(c1,c2); 
    } while (m1<m1alt);
    m1 = m1alt;
    if (m1>m) m1 = m; // m1: smallest positive secant slope
    if (xb>0) { c1 = mulu(m1,xb); b1 = subd(y1l,c1);} 
    else { c1 = muld(m1,xb); b1 = subd(y1l,c1);} // y=m1*x+b1;

// Calculaiton of the left secant equation:
    deltax = (xb-xa) / 100;  m2 = -1e306;
    SCHRANKE(xa,eps,y1l,y); 
    do {
	m1alt = m2;
	deltax *= 1.1;
	x = subd(xa,deltax);
	SCHRANKE(x,eps,y,y2l);
	c1 = subd(y1l,y2l);
	c2 = subd(xa,x);
	m2 = divd(c1,c2);
    } while (m2>=m1alt);
    deltax /= 1.21;   m = m1alt;
    m2 = -1e306;
    do {
	m1alt = m2;   deltax *= 1.001;
	x = subd(xa,deltax);
	SCHRANKE(x,eps,y,y2l);
	c1 = subd(y1l,y2l);
	c2 = subd(xa,x);
	m2 = divd(c1,c2);
    } while (m2>m1alt);
    m2 = m1alt;
    if (m2<m) m2 = m;
    if (xa>0) {c1 = muld(m2,xa); b2 = subd(y1l,c1);}
    else {c1 = mulu(m2,xa); b2 = subd(y1l,c1);} // y = m2*x + b2;
// Calculating the point of intersection of the two secants:
    y = subd(b2,b1);
    if (y>0) {c1 = subu(m1,m2);  y = divd(y,c1);}
    else {c1 = subd(m1,m2);  y = divd(y,c1);}
    // y: downrounded x-coordinate of the secant intersection point  
    c1  = muld(m1,y);
    y1l = addd(c1,b1); 
    // y1l: downrounded y-coordinate of the secant intersection point
    // it holds:   y1l <= f(x0);
    x1 = xa;
    x2 = xb;  // [x1,x2] = [xa,xb]: Inclusion of the minimum point x0
    SCHRANKE(xa,eps,y,y2l);
    SCHRANKE(xb,eps,y,m1);
    if (m1>y2l) y2l = m1;
    y2 = y2l;
    y1 = y1l; // [y1,y2]: Inclusion of f(x0) and of f([x1,x2]);
} // MIN_EINSCHL

int main()
{
    string str = "***********************************";
    string strleer = "    ";
    real xs, eps, x1, x2, y1, y2, ZP = 1e-2;
    interval xe,ye;
    cout << endl;
    cout << strleer+str+str << endl;
    cout << strleer+"*   "+
         "                                                                 *" 
         << endl;
    cout << strleer+"*   "+
         "This program calculates an inclusion of the minimum point  x0    *" 
         << endl;
    cout << strleer+"*   "+
         "and an inclusion of the exact value f(x0), if f is convex and    *"
         << endl;
    cout << strleer+"*   "+
         "has a relative minimum at x = x0.                                *"
         << endl;
    cout << strleer+"*   "+
         "An approximation of x0 and the relative error bound eps(f) of    *"
         << endl;
    cout << strleer+"*   "+
         "the convex funktion f(x) must be known.                          *"
         << endl;
    cout << strleer+"*   "+
         "f(x)  must be implemented in the source code of this program.    *"
         << endl;
    cout << strleer+"*   "+
         "                                                                 *" 
         << endl;
    cout << strleer+str+str << endl;

    cout << "    ****************************************** " << endl;
    cout << "    *   Minimum point approximation xs = ?   * " << endl;
    cout << "    ****************************************** " << endl;
    cin >> xs;

    cout << "    ************************************************ " << endl;
    cout << "    *   Relative error bound of f(x),  epsf = ?    * " << endl;
    cout << "    ************************************************ " << endl;
    cin >> RndUp >> eps;
    cout << endl;
    cout << "    **   Please wait some moments ...." << endl << endl;
    MIN_EINSCHL(eps,ZP,xs,x1,x2,y1,y2);
    xe = interval(x1,x2);  ye = interval(y1,y2);
    cout << SetPrecision(15,15) 
         << "    **   x0    is included in: " << xe << endl;

    cout << "    **   f(x0) is included in: " << ye << endl << endl; 
} // main

