/*
**  CXSC is a C++ library for eXtended Scientific Computing
**
**  Copyright (C) 1990-2000 Institut fuer Angewandte Mathematik,
**                          Universitaet Karlsruhe, Germany
**            (C) 2000-2001 Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany   
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*  Calculating absolute and relative error bounds for the basic operations
**  (+,-,*,/)  in only high accuracy. The relative error for the basic 
**  operations is assumed as:   eps_m = 2^(-52) = 2.22044604...E-16  if the 
**  operation results are normalized IEEE-numbers. 
**
**  Functions to calculate the absolute error are named: 
**     abs_addh1(...)   for the basic operation +
**     abs_subh1(...)   for the basic operation -
**     abs_mulh1(...)   for the basic operation *
**     abs_divh1(...)   for the basic operation / 
**  The last sign '1' stands for:  absolute error of the two operands.
**  The last sign '2' stands for:  absolute error of the first operand and
**                                 relative error of the second operand.
**  The last sign '3' stands for:  relative error of the first operand and
**                                 absolute error of the second operand.
**  The last sign '4' stands for:  relative error of the two operands.
**
**  Functions to calculate the relative error are named: 
**     rel_addh1(...)   for the basic operation +
**     rel_subh1(...)   for the basic operation -
**     rel_mulh1(...)   for the basic operation *
**     rel_divh1(...)   for the basic operation / 
**  The meaning of last sign is the same as by calculating the absolute error.
**  The sign 'h' stands for:  high accuracy 
**
**  Functions to calculate the absolute or relative error bounds for 
**  standard functions with erroneuos arguments are named:
**     abs_name_abs(), abs_name_rel(), rel_name_abs(), rel_name_rel(),
**  where 'name' is the symbol for the standard function like lnp1 or exp. 
**
**  Blomquist, 20.08.03;
*/

#ifndef abs_relh_hpp_included
#define abs_relh_hpp_included

#include <interval.hpp>
#include "hilfe.hpp"

namespace cxsc
{

int e_IEEE(const real& x);
// Liefert den Exponenten e(x) in den Darstellungen:
// Normalisierter Bereich:   x = (-1)^s * 1.m * 2^e(x);
// Denormalisierten Bereich: x = (-1)^s * 0.m * 2^(-1022)
// Null:                     0 = (-1)^s * 0.0...0 * 2^0

int z_trail(const real& x);
// Liefert die Anzahl aufeinanderfolgender Null-Bits am Ende der Mantisse
// von x.       x=0 --> 53;  x=1 --> 52;  x=1.5 --> 51;  x=pred(2) --> 0;
// Vorsicht: Funktioniert zunchst nur auf INTEL-Systemen!!

int z_lead(const real& x);
// Liefert die Anzahl aufeinanderfolgender Null-Bits am Anfang der Mantisse
// von x.    x=0 --> 0;  x=MinReal --> 0;  x=minreal --> 52;

int e_IS(const interval& x);
// Realisiert die Funktion eIS(X) von Seite 18 aus Preprint Nr. 98/5

//---------------------------------------------------------------------------
// ----------- Calculating the upper bounds of the absolute errors ----------
//---------------------------------------------------------------------------
// ------------------------ Delta(a), Delta(b) ------------------------------
// --------------------------------------------------------------------------
void abs_addh1(const interval& A, const real& dela,
               const interval& B, const real& delb,
               interval& R, real& del) throw(); // Returns the interval R=A+B
         // and an upper bound del of the absolute error by calculating (a+b)
         // on the machine with high accuracy, where a and b are erroneous
         // maschine values.

void abs_subh1(const interval& A, const real& dela,
               const interval& B, const real& delb,
               interval& R, real& del) throw(); // Returns the interval R=A-B
         // and an upper bound del of the absolute error by calculating (a-b)
         // on the machine with high accuracy, where a and b are erroneous
         // maschine values.

void abs_mulh1(const interval& A, const real& dela,
               const interval& B, const real& delb,
               interval& R, real& del) throw(); // Returns the interval R=A*B
         // and an upper bound del of the absolute error by calculating (a*b)
         // on the machine with high accuracy, where a and b are erroneous
         // maschine values.  

void abs_divh1(const interval& A, const real& dela,
               const interval& B, const real& delb,
               interval& R, real& del) throw(); // Returns the interval R=A/B
         // and an upper bound del of the absolute error by calculating (a/b)
         // on the machine with high accuracy, where a and b are erroneous
         // maschine values.

//---------------------------------------------------------------------------
// ------------------------- Delta(a), Eps(b) -------------------------------
// --------------------------------------------------------------------------

void abs_addh2(const interval& A, const real& dela,
               const interval& B, const real& epsb,
               interval& R, real& del) throw(); // Returns the interval R=A+B
         // and an upper bound del of the absolute error by calculating (a+b)
         // on the machine with high accuracy, where a and b are erroneous
         // maschine values.

void abs_subh2(const interval& A, const real& dela,
               const interval& B, const real& epsb,
               interval& R, real& del) throw(); // Returns the interval R=A-B
         // and an upper bound del of the absolute error by calculating (a-b)
         // on the machine with high accuracy, where a and b are erroneous
         // maschine values.

void abs_mulh2(const interval& A, const real& dela,
               const interval& B, const real& epsb,
               interval& R, real& del) throw(); // Returns the interval R=A*B
         // and an upper bound del of the absolute error by calculating (a*b)
         // on the machine with high accuracy, where a and b are erroneous
         // maschine values.  

void abs_divh2(const interval& A, const real& dela,
               const interval& B, const real& epsb,
               interval& R, real& del) throw(); // Returns the interval R=A/B
         // and an upper bound del of the absolute error by calculating (a/b)
         // on the machine with high accuracy, where a and b are erroneous
         // maschine values.

//---------------------------------------------------------------------------
// ------------------------- Eps(a), Delta(b) -------------------------------
// --------------------------------------------------------------------------

void abs_addh3(const interval& A, const real& epsa,
               const interval& B, const real& delb,
               interval& R, real& del) throw(); // Returns the interval R=A+B
         // and an upper bound del of the absolute error by calculating (a+b)
         // on the machine with high accuracy, where a and b are erroneous
         // maschine values.

void abs_subh3(const interval& A, const real& epsa,
               const interval& B, const real& delb,
               interval& R, real& del) throw(); // Returns the interval R=A-B
         // and an upper bound del of the absolute error by calculating (a-b)
         // on the machine with high accuracy, where a and b are erroneous
         // maschine values.

void abs_mulh3(const interval& A, const real& epsa,
               const interval& B, const real& delb,
               interval& R, real& del) throw(); // Returns the interval R=A*B
         // and an upper bound del of the absolute error by calculating (a*b)
         // on the machine with high accuracy, where a and b are erroneous
         // maschine values.  

void abs_divh3(const interval& A, const real& epsa,
               const interval& B, const real& delb,
               interval& R, real& del) throw(); // Returns the interval R=A/B
         // and an upper bound del of the absolute error by calculating (a/b)
         // on the machine with high accuracy, where a and b are erroneous
         // maschine values.

//---------------------------------------------------------------------------
// -------------------------- Eps(a), Eps(b) --------------------------------
// --------------------------------------------------------------------------

void abs_addh4(const interval& A, const real& epsa,
               const interval& B, const real& epsb,
               interval& R, real& del) throw(); // Returns the interval R=A+B
         // and an upper bound del of the absolute error by calculating (a+b)
         // on the machine with high accuracy, where a and b are erroneous
         // maschine values.

void abs_subh4(const interval& A, const real& epsa,
               const interval& B, const real& epsb,
               interval& R, real& del) throw(); // Returns the interval R=A-B
         // and an upper bound del of the absolute error by calculating (a-b)
         // on the machine with high accuracy, where a and b are erroneous
         // maschine values.

void abs_mulh4(const interval& A, const real& epsa,
               const interval& B, const real& epsb,
               interval& R, real& del) throw(); // Returns the interval R=A*B
         // and an upper bound del of the absolute error by calculating (a*b)
         // on the machine with high accuracy, where a and b are erroneous
         // maschine values.  

void abs_divh4(const interval& A, const real& epsa,
               const interval& B, const real& epsb,
               interval& R, real& del) throw(); // Returns the interval R=A/B
         // and an upper bound del of the absolute error by calculating (a/b)
         // on the machine with high accuracy, where a and b are erroneous
         // maschine values.


//---------------------------------------------------------------------------
// ----------- Calculating the upper bounds of the relative errors ----------
//---------------------------------------------------------------------------
// ------------------------ Delta(a), Delta(b) ------------------------------
// --------------------------------------------------------------------------
void rel_addh1(const interval& A, const real& dela,
               const interval& B, const real& delb,
               interval& R, real& eps) throw(); // Returns the interval R=A+B
         // and an upper bound eps of the relative error by calculating (a+b)
         // on the machine with high accuracy, where a and b are erroneous
         // maschine values.

void rel_subh1(const interval& A, const real& dela,
               const interval& B, const real& delb,
               interval& R, real& eps) throw(); // Returns the interval R=A-B
         // and an upper bound eps of the relative error by calculating (a-b)
         // on the machine with high accuracy, where a and b are erroneous
         // maschine values.

void rel_mulh1(const interval& A, const real& dela,
               const interval& B, const real& delb,
               interval& R, real& eps) throw(); // Returns the interval R=A*B
         // and an upper bound eps of the relative error by calculating (a*b)
         // on the machine with high accuracy, where a and b are erroneous
         // maschine values.  

void rel_divh1(const interval& A, const real& dela,
               const interval& B, const real& delb,
               interval& R, real& eps) throw(); // Returns the interval R=A/B
         // and an upper bound eps of the relative error by calculating (a/b)
         // on the machine with high accuracy, where a and b are erroneous
         // maschine values.

//---------------------------------------------------------------------------
// ------------------------- Delta(a), Eps(b) -------------------------------
// --------------------------------------------------------------------------

void rel_addh2(const interval& A, const real& dela,
               const interval& B, const real& epsb,
               interval& R, real& eps) throw(); // Returns the interval R=A+B
         // and an upper bound eps of the relative error by calculating (a+b)
         // on the machine with high accuracy, where a and b are erroneous
         // maschine values.

void rel_subh2(const interval& A, const real& dela,
               const interval& B, const real& epsb,
               interval& R, real& eps) throw(); // Returns the interval R=A-B
         // and an upper bound eps of the relative error by calculating (a-b)
         // on the machine with high accuracy, where a and b are erroneous
         // maschine values.

void rel_mulh2(const interval& A, const real& dela,
               const interval& B, const real& epsb,
               interval& R, real& eps) throw(); // Returns the interval R=A*B
         // and an upper bound eps of the relative error by calculating (a*b)
         // on the machine with high accuracy, where a and b are erroneous
         // maschine values.  

void rel_divh2(const interval& A, const real& dela,
               const interval& B, const real& epsb,
               interval& R, real& eps) throw(); // Returns the interval R=A/B
         // and an upper bound eps of the relative error by calculating (a/b)
         // on the machine with high accuracy, where a and b are erroneous
         // maschine values.

//---------------------------------------------------------------------------
// ------------------------- Eps(a), Delta(b) -------------------------------
// --------------------------------------------------------------------------

void rel_addh3(const interval& A, const real& epsa,
               const interval& B, const real& delb,
               interval& R, real& eps) throw();  // Returns the interval 
               // R=A+B and an upper bound eps of the relative error by 
               // calculating (a+b) on the machine with high accuracy, 
               // where a and b are erroneous maschine values.

void rel_subh3(const interval& A, const real& epsa,
               const interval& B, const real& delb,
               interval& R, real& eps) throw();  // Returns the interval 
               // R=A-B and an upper bound eps of the relative error by 
               // calculating (a-b) on the machine with high accuracy, 
               // where a and b are erroneous maschine values.

void rel_mulh3(const interval& A, const real& epsa,
               const interval& B, const real& delb,
               interval& R, real& eps) throw();  // Returns the interval 
               // R=A*B and an upper bound eps of the relative error by 
               // calculating (a*b) on the machine with high accuracy, 
               // where a and b are erroneous maschine values.

void rel_divh3(const interval& A, const real& epsa,
               const interval& B, const real& delb,
               interval& R, real& eps) throw();  // Returns the interval 
               // R=A/B and an upper bound eps of the relative error by 
               // calculating (a/b) on the machine with high accuracy, 
               // where a and b are erroneous maschine values.

//---------------------------------------------------------------------------
// -------------------------- Eps(a), Eps(b) --------------------------------
// --------------------------------------------------------------------------

void rel_addh4(const interval& A, const real& epsa,
               const interval& B, const real& epsb,
               interval& R, real& eps) throw(); // Returns the interval R=A+B
         // and an upper bound eps of the relative error by calculating (a+b)
         // on the machine with high accuracy, where a and b are erroneous
         // maschine values.

void rel_subh4(const interval& A, const real& epsa,
               const interval& B, const real& epsb,
               interval& R, real& eps) throw(); // Returns the interval R=A-B
         // and an upper bound eps of the relative error by calculating (a-b)
         // on the machine with high accuracy, where a and b are erroneous
         // maschine values.

void rel_mulh4(const interval& A, const real& epsa,
               const interval& B, const real& epsb,
               interval& R, real& eps) throw(); // Returns the interval R=A*B
         // and an upper bound eps of the relative error by calculating (a*b)
         // on the machine with high accuracy, where a and b are erroneous
         // maschine values.  

void rel_divh4(const interval& A, const real& epsa,
               const interval& B, const real& epsb,
               interval& R, real& eps) throw(); // Returns the interval R=A/B
         // and an upper bound eps of the relative error by calculating (a/b)
         // on the machine with high accuracy, where a and b are erroneous
         // maschine values.

//****************************************************************************
//*                   Error bounds of standard functions                     *
//****************************************************************************

//****************************************************************************
//*                              Square root                                 *
//****************************************************************************

    void abs_sqrt_abs(const interval& A, const real& dela,
		      interval& R, real& abs);  // Returns the interval
    // R=sqrt(A) and the absolute error of the sqrt-function. A is the
    // interval of the exact arguments, and dela is the absolute error 
    // bound of the possibly defective function arguments.

    void abs_sqrt_rel(const interval& A, const real& epsa,
		      interval& R, real& abs);  // Returns the interval
    // R=sqrt(A) and the absolute error of the sqrt-function. A is the
    // interval of the exact arguments, and epsa is the relative error 
    // bound of the possibly defective function arguments.

    void rel_sqrt_abs(const interval& A, const real& dela,
		      interval& R, real& rel);  // Returns the interval
    // R=sqrt(A) and the relative error of the sqrt-function. A is the
    // interval of the exact arguments, and dela is the absolute error 
    // bound of the possibly defective function arguments.

    void rel_sqrt_rel(const interval& A, const real& epsa,
		      interval& R, real& rel);  // Returns the interval
    // R=sqrt(A) and the relative error of the sqrt-function. A is the
    // interval of the exact arguments, and epsa is the relative error 
    // bound of the possibly defective function arguments. 

//****************************************************************************
//*                         Exponential function                             *
//****************************************************************************

    void abs_exp_abs(const interval& A, const real& dela,
		     interval& R, real& abs);  // Returns the interval
    // R=exp(A) and the absolute error of the exp-function. A is the
    // interval of the exact arguments, and dela is the absolute error 
    // bound of the possibly defective function arguments. 

    void abs_exp_rel(const interval& A, const real& epsa,
		     interval& R, real& abs);  // Returns the interval
    // R=exp(A) and the absolute error of the exp-function.   A is the
    // interval of the exact arguments, and epsa is the relative error 
    // bound of the possibly defective function arguments.

    void rel_exp_abs(const interval& A, const real& dela,
		     interval& R, real& rel);  // Returns the interval
    // R=exp(A) and the relative error of the exp-function.   A is the
    // interval of the exact arguments, and dela is the absolute error 
    // bound of the possibly defective function arguments.

    void rel_exp_rel(const interval& A, const real& epsa,
		     interval& R, real& rel);  // Returns the interval
    // R = exp(A) and the relative error of the exp-function. A is the
    // interval of the exact arguments, and epsa is the relative error 
    // bound of the possibly defective function arguments.

//****************************************************************************
//*                           Logarithm function                             *
//****************************************************************************

    void abs_ln_abs(const interval& A, const real& dela,
		    interval& R, real& abs);  // Returns the interval
    // R=ln(A) and the absolute error of the ln-function.    A is the
    // interval of the exact arguments, and dela is the absolute error 
    // bound of the possibly defective function arguments.

    void abs_ln_rel(const interval& A, const real& epsa,
		    interval& R, real& abs);   // Returns the interval
    // R=ln(A) and the absolute error of the ln-function.     A is the
    // interval of the exact arguments, and epsa is the relative error 
    // bound of the possibly defective function arguments.

    void rel_ln_abs(const interval& A, const real& dela,
		    interval& R, real& rel);  // Returns the interval
    // R=ln(A) and the relative error of the ln-function.    A is the
    // interval of the exact arguments, and dela is the absolute error 
    // bound of the possibly defective function arguments.

    void rel_ln_rel(const interval& A, const real& epsa,
		    interval& R, real& rel);  // Returns the interval
    // R = ln(A) and the relative error of the ln-function.  A is the
    // interval of the exact arguments, and epsa is the relative error 
    // bound of the possibly defective function arguments.

//****************************************************************************
//*                           function Ln(1+a)                               *
//****************************************************************************

    void abs_lnp1_abs(const interval& A, const real& dela,
		      interval& R, real& abs); // Returns the interval
    // R=ln(1+A) and the absolute error of the ln-function.   A is the
    // interval of the exact arguments, and dela is the absolute error 
    // bound of the possibly defective function arguments.

    void abs_lnp1_rel(const interval& A, const real& epsa,
		      interval& R, real& abs); // Returns the interval
    // R=ln(1+A) and the absolute error of the ln-function.   A is the
    // interval of the exact arguments, and epsa is the relative error 
    // bound of the possibly defective function arguments.

    void rel_lnp1_abs(const interval& A, const real& dela,
		      interval& R, real& rel); // Returns the interval
    // R=ln(1+A) and the relative error of the ln-function.   A is the
    // interval of the exact arguments, and dela is the absolute error 
    // bound of the possibly defective function arguments.

    void rel_lnp1_rel(const interval& A, const real& epsa,
		      interval& R, real& rel); // Returns the interval
    // R = ln(1+A) and the relative error of the ln-function. A is the
    // interval of the exact arguments, and epsa is the relative error 
    // bound of the possibly defective function arguments.

//****************************************************************************
//*                           function acosh(a)                              *
//****************************************************************************

void rel_acosh_abs(const interval& A, const real& dela,
		   interval& R, real& rel); // Returns the interval
    // R=acosh(A) and the relative error of the acosh-function. A is the
    // interval of the exact arguments, and dela is the absolute error 
    // bound of the possibly defective function arguments.

void rel_acosh_rel(const interval& A, const real& epsa,
		   interval& R, real& rel); // Returns the interval
    // R=acosh(A) and the relative error of the acosh-function. A is the
    // interval of the exact arguments, and epsa is the relative error 
    // bound of the possibly defective function arguments.


//****************************************************************************
//*                Some functions for a simple error calculus                *
//****************************************************************************

real eps_2(const real& eps_a, const real& eps_b);
    // returns an upperbound for |eps_ab| with 
    // (1+eps_a)*(1+eps_b) = 1+eps_ab

real eps_3(const real& eps_a, const real& eps_b, const real& eps_k);
    // returns an upperbound for |eps_abk| with 
    // (1+eps_a)*(1+eps_b)*(1+eps_k) = 1+eps_abk

real eps_4(const real& eps_a, const real& eps_b,
           const real& eps_c, const real& eps_d);
    // returns an upperbound for |eps_abcd| with 
    // (1+eps_a)*(1+eps_b)*(1+eps_c)*(1+eps_c) = 1+eps_abcd

real eps_pow(const real& eps, int n) throw();
    // returns an upperbound for |eps_p| with 
    // (1+eps)^n = (1+eps_p), n >= 1;

real eps_rezi(const real& eps) throw();
    // 1/(1+eps) = 1+eps_1; eps<=0.5 ! 

real eps_ZdN(const real& eps_Z, const real& eps_N);
    // Error bound of a fraction; 
    // The rel. error bound for the division is used with eps_h = 2^{-52} 
    // for only high accuracy! 

} // namespace cxsc 

#endif













