/*
**  CXSC is a C++ library for eXtended Scientific Computing
**
**  Copyright (C) 1990-2000 Institut fuer Angewandte Mathematik,
**                          Universitaet Karlsruhe, Germany
**            (C) 2000-2001 Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany   
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "bnd_arh.hpp"
#include <dot.hpp>
#include <interval.hpp>

using namespace std;
namespace cxsc {
// ---------------------------------------------------------------------------
// ------------------------------ T(x) = 1+x ---------------------------------
// ---------------------------------------------------------------------------

static const real c54  = comp(0.5,55);   // 2^(+54)
static const real c53  = comp(0.5,54);   // 2^(+53)
static const real c_52 = comp(0.5,-51);  // 2^(-52)
static const real c_53 = comp(0.5,-52);  // 2^(-53)
static const real c_54 = comp(0.5,-53);  // 2^(-54)
static const real c_4  = comp(0.5,-3);   // 2^(-4) = 1/16
static const real c_1 = 1.5*c_54;
static const real c_2 = 1.75*c_54;

static real aerr_xp1h1(const interval& x)
// The upper bound of the absolute error by calculating (1+r) on the machine
// with r in [-MaxReal,-2^53] is returned. Only high accuracy is assumed!
{
    real ix = Inf(x), y = 0;
    if (ix<-c54) 
    {
	y = addd(ix,1);
	y = abs(y);
	times2pown(y,-52); // y = 2^(-52)*|ix+1|
    }  // ix >= -c54
    else if (ix<-c53) y = 1;
    return y;
}

static real aerr_xp1h2(const interval& x)
// The upper bound of the absolute error by calculating (1+r) on the machine
// with r in [-2^53,0.5] is returned. Only high accuracy is assumed!
{
    real ix = Inf(x), sx = Sup(x),y;
    
    if (sx>=minreal) return c_52;
    // sx <= 0
    if (ix>=-minreal)
    {
	y = (ix==0) ? 0 : c_53;
	return y;
    }
    if (sx>-c_4)     return c_53;
    if (sx>-0.125)   return c_2;
    if (sx>-0.25)    return c_1;
    if (sx>-0.5)     return c_54;
    return real(0.0);
}

static real aerr_xp1h3(const interval& x)
// The upper bound of the absolute error by calculating (1+r) on the machine
// with r in [0.5,2] is returned. Only high accuracy is assumed!
{
    real ix = Inf(x), sx = Sup(x), y;

    y = 0.0;
    if (sx<1) y = c_53;
    else if (ix<2) y = c_52;
    return y;
}

static real aerr_xp1h4(const interval& x)
// The upper bound of the absolute error by calculating (1+r) on the machine
// with r in [2,Maxreal] is returned. Only high accuracy is assumed!
{
    real ix = Inf(x), sx = Sup(x), y(0);
    int exsx = expo(sx);
    if (exsx<=53) 
    {
	if (sx > comp(0.5,exsx+1)-1) y = comp(0.5,exsx-52);
	else if (ix<comp(0.5,exsx))  y = comp(0.5,exsx-53);
    }
    else
    {
	y = comp(0.5,exsx-52);
	y = subu(y,1);
    }
    return y;
}

real Rnd_1pxh(const interval& x) throw()
// Liefert Oberschranke Del des absoluten Fehlers bei der nur hochgenauen 
// Berechnung von 1+r auf der Maschine; r ist Element aus x.
// Ausfuehrlich getestet!!
{
    real t, y=0;

    interval z1,z2,z3,z4;
    z1 = interval(-MaxReal,-c53);
    z2 = interval(-c53,0.5);
    z3 = interval(0.5,2);
    z4 = interval(2,MaxReal);

    if (!Disjoint(x,z1)) 
    {
	t = aerr_xp1h1(x & z1);
	if (t>y) y = t;
    }
    if (!Disjoint(x,z2)) 
    {
	t = aerr_xp1h2(x & z2);
	if (t>y) y = t;
    }
    if (!Disjoint(x,z3)) 
    {
	t = aerr_xp1h3(x & z3);
	if (t>y) y = t;
    }
    if (!Disjoint(x,z4)) 
    {
	t = aerr_xp1h4(x & z4);
	if (t>y) y = t;
    }
    return y;
} // Rnd_1pxh

void abs_1px_delh(const interval& X, const real& del,
                        interval& T, real& bnd) throw()
// Berechnung des abs. Gesamtfehlers, wenn T(x)=1+x in nur hochgenauer
// Arithmetik ausgewertet wird. X ist das Intervall der exakten
// Werte, und del ist die abs. Fehlerschr. der gestrten Maschinenwerte.
// T = 1+X ist eine Einschlieung der exakten Werte T(x)=1+x. 
{
    real d = abs(del);
    T = 1+X;  // T: Einschlieung der exakten T(x)-Werte
    interval Xt = X+interval(-d,d);  // Xt: Einschlieung der
                                     // gestrten Maschinenzahlen
    bnd = Rnd_1pxh(Xt);  // bnd: Absoluter Rundungsfehler
    bnd = addu(d,bnd);   // bnd: Absoluter Gesamtfehler
} // abs_1px_delh

void abs_1px_epsh(const interval& X, const real& eps,
                        interval& T, real& bnd) throw()
// Berechnung des abs. Gesamtfehlers, wenn T(x)=1+x in nur hochgenauer
// Arithmetik ausgewertet wird. X ist das Intervall der exakten
// Werte, und eps ist die rel. Fehlerschr. der gestrten Maschinenwerte.
// T = 1+X ist eine Einschlieung der exakten Werte T(x)=1+x.     
{
    T = 1+X;  // T: Einschlieung der exakten T(x)-Werte
    real d = abs(eps);
    d = mulu(d,Sup(abs(X)));  // d: abs. Fehlerschr.     
    interval Xt = X+interval(-d,d);  // Xt: Einschlieung der
                                     // gestrten Maschinenzahlen
    bnd = Rnd_1pxh(Xt); // bnd: Absoluter Rundungsfehler
    bnd = addu(d,bnd);  // bnd: Absoluter Gesamtfehler
} // abs_1px_epsh

void rel_1px_delh(const interval& X, const real& del,
                        interval& T, real& bnd) throw()
// Berechnung des rel. Gesamtfehlers, wenn T(x)=1+x in nur hochgenauer
// Arithmetik ausgewertet wird. X ist das Intervall der exakten
// Werte, und del ist die abs. Fehlerschr. der gestrten Maschinenwerte.
// T = 1+X ist eine Einschlieung der exakten Werte T(x)=1+x.
// Bei zu breitem X ist eine Intervallunterteilung notwendig! 
{
    real d = abs(del);
    T = 1+X;  // T: Einschlieung der exakten T(x)-Werte
    interval Xt = X+interval(-d,d);  // Xt: Einschlieung der
                                     // gestrten Maschinenzahlen
    bnd = Rnd_1pxh(Xt); // bnd: Absoluter Rundungsfehler
    bnd = addu(d,bnd);  // bnd: Absoluter Gesamtfehler
    if (sign(bnd)!=0) 
    {
	d = Inf( abs(T) );
	if (sign(d)==0)
	{
	    cerr << "rel_1px_delm(): Relative error doesn't exist!" << endl;
	    exit(1);
	}
        else bnd = divu(bnd,d); // bnd: Relativer Gesamtfehler
    }
} // rel_1px_delh

void rel_1px_epsh(const interval& X, const real& eps,
                        interval& T, real& bnd) throw()
// Berechnung des rel. Gesamtfehlers, wenn T(x)=1+x in nur hochgenauer
// Arithmetik ausgewertet wird. X ist das Intervall der exakten
// Werte, und eps ist die rel. Fehlerschr. der gestrten Maschinenwerte.
// T = 1+X ist eine Einschlieung der exakten Werte T(x)=1+x. 
// Bei zu breitem X ist eine Intervallunterteilung notwendig!    
{
    T = 1+X;  // T: Einschlieung der exakten T(x)-Werte
    real d = abs(eps);
    d = mulu(d,Sup(abs(X)));  // d: abs. Fehlerschr.     
    interval Xt = X+interval(-d,d);  // Xt: Einschlieung der
                                     // gestrten Maschinenzahlen
    bnd = Rnd_1pxh(Xt); // bnd: Absoluter Rundungsfehler
    bnd = addu(d,bnd);  // bnd: Absoluter Gesamtfehler
    if (sign(bnd)!=0) 
    {
	d = Inf( abs(T) );
	if (sign(d)==0)
	{
	    cerr << "rel_1px_epsm(): Relative error doesn't exist!" << endl;
	    exit(1);
	}
        else bnd = divu(bnd,d); // bnd: Rel. error calculated
    }
} // rel_1px_epsh

// ---------------------------------------------------------------------------
// ---------------------------- T(x) = 1-x -----------------------------------
// ---------------------------------------------------------------------------

void abs_1mx_delh(const interval& X, const real& del,
                        interval& T, real& bnd) throw()
// Berechnung des abs. Gesamtfehlers, wenn T(x)=1-x in nur hochgenauer
// Arithmetik ausgewertet wird. X ist das Intervall der exakten
// Werte, und del ist die abs. Fehlerschr. der gestrten Maschinenwerte.
// T = 1-X ist eine Einschlieung der exakten Werte T(x) = 1-x = 1+(-x). 
{   
    abs_1px_delh(-X,del,T,bnd);
} // abs_1mx_delh

void abs_1mx_epsh(const interval& X, const real& eps,
                        interval& T, real& bnd) throw()
// Berechnung des abs. Gesamtfehlers, wenn T(x)=1-x in nur hochgenauer
// Arithmetik ausgewertet wird. X ist das Intervall der exakten
// Werte, und eps ist die rel. Fehlerschr. der gestrten Maschinenwerte.
// T = 1-X ist eine Einschlieung der exakten Werte T(x) = 1-x = 1+(-x). 
{   
    abs_1px_epsh(-X,eps,T,bnd);
} // abs_1mx_epsh

void rel_1mx_delh(const interval& X, const real& del,
                        interval& T, real& bnd) throw()
// Berechnung des rel. Gesamtfehlers, wenn T(x)=1-x in nur hochgenauer
// Arithmetik ausgewertet wird. X ist das Intervall der exakten
// Werte, und del ist die abs. Fehlerschr. der gestrten Maschinenwerte.
// T = 1-X ist eine Einschlieung der exakten Werte T(x) = 1-x = 1+(-x). 
{   
    rel_1px_delh(-X,del,T,bnd);
} // abs_1mx_delh

void rel_1mx_epsh(const interval& X, const real& eps,
                        interval& T, real& bnd) throw()
// Berechnung des rel. Gesamtfehlers, wenn T(x)=1-x in nur hochgenauer
// Arithmetik ausgewertet wird. X ist das Intervall der exakten
// Werte, und eps ist die rel. Fehlerschr. der gestrten Maschinenwerte.
// T = 1-X ist eine Einschlieung der exakten Werte T(x) = 1-x = 1+(-x). 
{   
    rel_1px_epsh(-X,eps,T,bnd);
} // rel_1mx_epsh

// ---------------------------------------------------------------------------
// ---------------------------- T(x) = x-1 -----------------------------------
// ---------------------------------------------------------------------------

void abs_xm1_delh(const interval& X, const real& del,
                        interval& T, real& bnd) throw()
// Berechnung des abs. Gesamtfehlers, wenn T(x)=x-1 in nur hochgenauer
// Arithmetik ausgewertet wird. X ist das Intervall der exakten
// Werte, und del ist die abs. Fehlerschr. der gestrten Maschinenwerte.
// T = X-1 ist eine Einschlieung der exakten Werte T(x)=x-1 = -(1+(-x)). 
{   
    abs_1px_delh(-X,del,T,bnd);
    T = -T;
} // abs_xm1_delh

void abs_xm1_epsh(const interval& X, const real& eps,
                        interval& T, real& bnd) throw()
// Berechnung des abs. Gesamtfehlers, wenn T(x)=x-1 in nur hochgenauer
// Arithmetik ausgewertet wird. X ist das Intervall der exakten
// Werte, und eps ist die rel. Fehlerschr. der gestrten Maschinenwerte.
// T = X-1 ist eine Einschlieung der exakten Werte T(x)=x-1 = -(1+(-x)). 
{   
    abs_1px_epsh(-X,eps,T,bnd);
    T = -T;
} // abs_xm1_epsh

void rel_xm1_delh(const interval& X, const real& del,
                        interval& T, real& bnd) throw()
// Berechnung des rel. Gesamtfehlers, wenn T(x)=x-1 in nur hochgenauer
// Arithmetik ausgewertet wird. X ist das Intervall der exakten
// Werte, und del ist die abs. Fehlerschr. der gestrten Maschinenwerte.
// T = X-1 ist eine Einschlieung der exakten Werte T(x)=x-1 = -(1+(-x)). 
{   
    rel_1px_delh(-X,del,T,bnd);
    T = -T;
} // abs_xm1_delh

void rel_xm1_epsh(const interval& X, const real& eps,
                        interval& T, real& bnd) throw()
// Berechnung des rel. Gesamtfehlers, wenn T(x)=x-1 in nur hochgenauer
// Arithmetik ausgewertet wird. X ist das Intervall der exakten
// Werte, und eps ist die rel. Fehlerschr. der gestrten Maschinenwerte.
// T = X-1 ist eine Einschlieung der exakten Werte T(x)=x-1 = -(1+(-x)). 
{   
    rel_1px_epsh(-X,eps,T,bnd);
    T = -T;
} // rel_xm1_epsh


//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

real abs_err_xpah(const real& x, const real& a) throw()
// Berechnet mit Akku garantierte Oberschranke des absoluten Rundungs-
// fehlers, wenn (x+a) auf der Maschine nur hochgenau ausgewertet wird.
// Ist die exakte Summe x+a keine Maschinenzahl und sind y1,y2 die 
// einschliessenden Maschinenzahlen von x+a, so wird zu dem yi gerundet, 
// mit dem der absolute Fehler maximal wird.  
{
    dotprecision dot,dot1;
    interval z;
    real xpa,delx1,delx2;
    dot = a;
    dot += x;    // dot = a+x; exact calculation!
    xpa = addu(x,a); // rounding upwards
    dot1 = dot - xpa;  // dot1 = delx1;
    if (sign(dot1)==0) delx1 = 0;
    else 
    {
	rnd(dot1,z);
	z = abs(z);
	delx1 = Sup(z);
    }
    xpa = addd(x,a);   // rounding downwards
    dot1 = dot - xpa;  // dot1 = delx2;
    if (sign(dot1)==0) delx2 = 0;
    else 
    {
	rnd(dot1,z);
	z = abs(z);
	delx2 = Sup(z);
    }
    if (delx2>delx1) delx1 = delx2;
    return delx1;
} // abs_err_xpah


} // namespace cxsc













