/*
**  CXSC is a C++ library for eXtended Scientific Computing
**
**  Copyright (C) 1990-2000 Institut fuer Angewandte Mathematik,
**                          Universitaet Karlsruhe, Germany
**            (C) 2000-2001 Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany   
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef bnd_util_hpp_included
#define bnd_util_hpp_included

#include "abs_relh.hpp"  // Wegen rel_mulh4()
#include "bnd_arh.hpp"   // Wegen rel_1px_epsh()

#include <l_interval.hpp>
#include <intvector.hpp>
#include <ivector.hpp>

namespace cxsc
{

void Max_bnd_Xi(real (*FPt)(const interval& Xi, const real& bndx), 
                                   // FPt: Funktions-Zeiger auf Funktion mit 
                                   // dem Rckgabetyp real.
                const interval& X, // X: Zu unterteilendes Intervall
	        const real& bndx,  // bndx: Fehlerschranke der Elemente aus X
                const real& diam,  // diam: Mantissen-Durchmesser
	              real& bnd);  // bnd: Maximum aller Fehlerschranken
// Max_bnd_Xi zerlegt X in Teilintervalle Xi mit dem Mantissendurchmesser 
// diam; diam = 1E-5 ist meist eine gute Wahl. Fr jedes Teilintervall Xi
// wird mit der Funktion FPt eine Fehlerschranke berechnet, wobei bndx die
// Fehlerschranke der gestrten Elemente aus Xi sind. Der Referenzparameter
// bnd erhlt das Maximum aller Fehlerschranken ber den einzelnen Xi.

void Max_Rat_Xi(real (*FPt)(const interval&, const real&, 
                            const ivector&, const ivector&, const int&,
                            const real&),    // Funktions-Zeiger
                const interval& X,   // X: Zu unterteilendes Intervall der
		                     //    exakten Argumente
		const real& bndx,    // bndx: abs. oder rel. Fehlerschranke
                                     //       der gestrten Argumente
                const real& diam,    // diam: Mantissen-Durchmesser bei der
                                     //       Intervallunterteilung
		const ivector& ia,   // Optimale Koeffizienteneinschlieung
                const ivector& ib,   // durch Punktintervalle
		const int& k,        // k:    u(x) = x^k
		const real& x0,      // x0:   Entwicklingspunkt
	              real& bnd);    // bnd: Maximum aller Fehlerschranken.
// Max_Rat_Xi unterteilt X mittels diam in Teilintervalle Xi, berechnet fr
// alle Xi mit der Funktion FPt() die gesuchten Fehlerschranken und liefert 
// mit bnd das Maximum dieser Fehlerschranken.
// Bedeutung der Funktionsparameter von FPt:
// 1. Param.: Teilintervall der exakten Argumente fr die Fehlerabschtzung;
//            Dieses Teilintervall wird in dieser Funktion als Xi berechnet.
// 2. Param.: Absolute oder rel. Fehlerschranke der gestrten Argumente.
//            Dieser Parameter wird beim Aufruf von FPt durch bndx ersetzt.
// 3. Param.: Intervallvektor, mit den optimalen Koeffztn.-Einschlieungen.
//            Dieser Parameter wird beim Aufruf von FPt durch  ia  ersetzt.
// 4. Param.: Intervallvektor, mit den optimalen Koeffztn.-Einschlieungen.
//            Dieser Parameter wird beim Aufruf von FPt durch  ib  ersetzt.
// 5. Param.: Integer-Gre k bezglich u(x) = x^k. 
//            Dieser Parameter wird beim Aufruf von FPt durch  k  ersetzt. 
// 6. Param.: real-Gre x0 ist Entwicklungspunkt

void eps2fractions(const string& eps, 
		   real& Z1p, real& N1p, real& Z1m, real& N1m);
// For a given relative error bound eps this function delivers the
// numerators and the denominators for the terms 1+eps and 1-eps.
// Example:  With eps = "2.97768e-15" we get:
// 1-eps = Z1m / N1m = 9007199254740962.0 / 9007199254740992.0
// 1+eps = Z1p / N1p = 4503599627370513.0 / 4503599627370496.0

void IEEE2frac_int(const real& x, real& z, real& n);
// For a given IEEE-number x the two whole-numbered IEEE-variables
// z,n are calculated, so that x = z/n;
// Notice, that the machine division z/n will be performed errorfree, 
// because the exact quotient z/n is the IEEE-number x. 
// The programm output of the results z,n should be done with:
// cout << SetPrecision(1,1) << Fixed << "z = " << z << endl;
// cout << "n = " << n << endl;

void frac_tools(const int& Z, const int& N,
		interval& x, real& Next, real& del, real& eps);
// Z is the numerator and N is the denominator of the fraction Z/N;
// x is the optimal inclusion of Z/N in the IEEE double format;
// Next is the nearest real number of the fraction Z/N;
// del is an upper bound of the absolute error |Z/N - Next|
// eps is an upper bound of the relative error |Z/N - Next|/|Z/N|

void Polynomgrad(int& N, int& k);
// Interaktive Eingabe von:
// N: Polynomgrad
// k: Exponent bezglich u=x^k, 1<=k<=6;

void Koeff_Rat(intvector& Za, intvector& Ne);
// Der Anwender wird aufgefordert, in die Vektoren Za,Ne die ganzzahligen
// Werte fr Zhler und Nenner der Polynomkoeffizienten einzulesen.
// Vorausgesetzt wird:  a[0]=1;  
// Lb(Za)=Lb(Ne)=1;  Polynomgrad = Ub(Za) = Ub(Ne) >= 1;

void Term_Dat(interval& X, real& diam);
// Interaktive Eingabe von:
// X   : Intervall der Term-Maschinen-Agumente
// diam: Mit diam wird die Intervallunterteilung gesteuert

void Poly_Dat_rel(interval& X, real& epsx, real& diam);
// Interaktive Eingabe von:
// X:    Intervall der exakten Polynomargumente
// epsx: Relativer Fehler der gesrten Polynomargumente
// diam: Mit diam wird die Intervallunterteilung gesteuert

real Horn_relh1(const intvector& Za, const intvector& Ne, const interval& X,
                const real& epsx, const int& k, const real& diam);
// Liefert mit bnd eine Oberschranke des relativen Fehlers, wenn das Polynom
// P_N(x) = 1 + a[1]*u^1 +...+a[N]*u^N,  u=x^k, 1<= k <=6,  nach dem Horner-
// schema auszuwerten ist. Die Vektoren Za und Ne, mit Ub(Za)=Ub(Ne)=N, mssen 
// fr jeden Koeffizienten a[j], j=1,..,N, die entsprechenden integer-Zahlen 
// fr Zhler und Nenner enthalten. Die exakten rationalen Koeffizienten
// a[j]=Za[j]/Ne[j] werden mit der Funktion Koeff_tools() zur jeweils nchsten
// Rasterzahl Next[j] aus S(2,53) gerundet, und mit diesen Koeffizienten
// Next[j] ist dann das Polynom nach dem Hornerschema in nur hochgenauer
// Arithmetik auszuwerten.
// X ist das Intervall der exakten Polynomargumente x aus X. Wurden die 
// Polynomargumente fehlerhaft berechnet, so ist epsx die entsprechende 
// relative Fehlerschranke. 
// X wird in Teilintervalle Xi zerlegt, um bnd zu minimieren; mit diam wird
// der Durchmesser der X1 gesteuert. diam = 10^(-5) ist i.a. eine gute Wahl.

real Horn_relh(const rvector& a, const real& x0, const interval& X,
               const real& epsx, const int& k, const real& diam);
// Liefert mit bnd eine Oberschranke des relativen Fehlers, wenn das Polynom
// P_N(u) = a[0] + a[1]*u^1 +...+a[N]*u^N,  u=(x-x0)^k, 1<= k <=6,  nach dem
// Hornerschema auszuwerten ist. 
// a ist der Vektor der exakten Polynomkoeffizienten a[j], j=0,1,...,N;
// x0 ist der Polynom-Entwicklungspunkt; x0 in S(2,53) muss erfllt sein!
// X ist das Intervall der exakten Polynomargumente x aus X. Wurden die 
// Polynomargumente fehlerhaft berechnet, so ist epsx die entsprechende 
// relative Fehlerschranke.
// X wird in Teilintervalle Xi zerlegt, um bnd zu minimieren; mit diam wird
// der Durchmesser der X1 gesteuert. diam = 10^(-5) ist i.a. eine gute Wahl.

real Rat_relh(const rvector& a, const rvector& b, const real& x0, 
              const interval& X, const real& epsx, const int& k, 
              const real& diam);
// Liefert mit bnd eine Oberschranke des relativen Fehlers, wenn die
// rationale Funktion   R(u)=P_N(u)/Q_M(u), u=(x-x0)^k, 1<=k<=6, mit den
// Polynomen:  P_N(u) = a[0] + a[1]*u^1 +...+a[N]*u^N,  
//             Q_M(u) = b[0] + b[1]*u^1 +...+b[M]*u^M,
// nach dem Hornerschema ausgewertet wird. 
// a ist der Vektoren der exakten Koeffizienten a[j], j=0,1,...,N;
// b ist der Vektoren der exakten Koeffizienten b[j], j=0,1,...,M;
// x0 ist der Polynom-Entwicklungspunkt; x0 in S(2,53) muss erfllt sein!
// X ist das Intervall der exakten Polynomargumente x aus X. Wurden die 
// Polynomargumente fehlerhaft berechnet, so ist epsx die entsprechende 
// relative Fehlerschranke.
// X wird in Teilintervalle Xi zerlegt, um bnd zu minimieren; mit diam wird
// der Durchmesser der X1 gesteuert. diam = 10^(-5) ist i.a. eine gute Wahl.

real Term_relh(real (*FPt)(const interval& Xi),
               const interval& X, const real& diam);
// Liefert mit bnd eine Oberschranke des relativen Fehlers, wenn ein
// gegebener Term auszuwerten ist. Die Berechnung der relativen Fehler-
// schranke muss mit der Funktion FPt() erfolgen.
// X ist das Intervall der exakten Termargumente x aus X.
// X wird in Teilintervalle Xi zerlegt, um bnd zu minimieren; mit diam wird
// der Durchmesser der Xi gesteuert. diam = 10^(-5) ist i.a. eine gute Wahl.

} // namespace cxsc 

#endif













